import chromadb
from chromadb.api.models.Collection import Collection
from openai import OpenAI
from typing import List, Optional, Dict
from chromadb.utils.embedding_functions import OpenAIEmbeddingFunction
import uuid
import time
import os
import asyncio
import logging
from collections import defaultdict

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class ChromaClientFactory:
    _client = None

    @classmethod
    def get_client(cls):
        if cls._client is not None:
            return cls._client
        env = os.getenv("APP_ENV", "production").lower()
        if env == "test":
            cls._client = chromadb.EphemeralClient()
        else:
            persist_dir = os.getenv("CHROMA_PERSIST_DIR", "./chroma_data")
            cls._client = chromadb.PersistentClient(path=persist_dir)
        return cls._client

class LongMemory:
    def __init__(self, collection_name: str = "long_term_memory", openai_api_key: str = "...", ttl_seconds: Optional[int] = None, client=None):
        self.openai = OpenAI(api_key=openai_api_key)
        self.embedding_fn = OpenAIEmbeddingFunction(api_key=openai_api_key, model_name="text-embedding-3-small")
        self.client = client or ChromaClientFactory.get_client()
        self.collection: Collection = self.client.get_or_create_collection(
            name=collection_name,
            metadata={"hnsw:space": "cosine", "hnsw:M": 32, "hnsw:construction_ef": 128},
            embedding_function=self.embedding_fn
        )
        self.ttl_seconds = ttl_seconds
        self._locks = defaultdict(asyncio.Lock)

    def _log_exception(self, message: str, exc: Exception):
        logger.exception(f"[LongMemory] {message}: {type(exc).__name__} - {exc}")

    def _adaptive_top_k(self, text: str, max_k: int = 5) -> int:
        length = len(text)
        if length < 100:
            return 1
        elif length < 300:
            return 3
        return max_k

    def _current_timestamp(self) -> float:
        return time.time()

    def _is_expired(self, last_used: float, threshold: float) -> bool:
        return last_used < threshold

    async def query_by_metadata(self, filter: Dict, text: str, top_k: int = 3) -> List[dict]:
        logger.debug(f"[LongMemory][query_by_metadata] получил аргументы: filter={filter}, text={text}, top_k={top_k}")
        try:
            adaptive_k = self._adaptive_top_k(text, top_k)
            results = self.collection.query(query_texts=[text], n_results=adaptive_k, where=filter)
            for i in range(len(results["ids"][0])):
                await self._touch_record(results["ids"][0][i])
            output = [
                {
                    "id": results["ids"][0][i],
                    "text": results["documents"][0][i],
                    "metadata": results["metadatas"][0][i],
                    "distance": results["distances"][0][i],
                }
                for i in range(len(results["documents"][0]))
                if not self._is_expired(results["metadatas"][0][i].get("_last_used", 0), self._current_timestamp() - (self.ttl_seconds or 0))
            ]
            logger.debug(f"[LongMemory][query_by_metadata] вернул результат: {output}")
            return output
        except Exception as e:
            self._log_exception("query_by_metadata failed", e)
            return []

    async def _touch_record(self, record_id: str):
        logger.debug(f"[LongMemory][_touch_record] обновление _last_used для записи: {record_id}")
        lock = self._locks[record_id]
        async with lock:
            try:
                record = self.collection.get(ids=[record_id])
                if record["metadatas"]:
                    metadata = record["metadatas"][0]
                    metadata["_last_used"] = self._current_timestamp()
                    text = record["documents"][0]
                    self.collection.update(ids=[record_id], documents=[text], metadatas=[metadata])
            except Exception as e:
                self._log_exception(f"_touch_record failed for {record_id}", e)

    async def add_record(self, text: str, metadata: Optional[Dict] = None, record_id: Optional[str] = None) -> str:
        logger.debug(f"[LongMemory][add_record] получил аргументы: text={text}, record_id={record_id}, metadata={metadata}")
        record_id = record_id or str(uuid.uuid4())
        metadata = metadata or {}
        metadata["_created"] = self._current_timestamp()
        metadata["_last_used"] = metadata["_created"]
        try:
            self.collection.add(documents=[text], ids=[record_id], metadatas=[metadata])
            logger.debug(f"[LongMemory][add_record] вернул результат: {record_id}")
            logger.debug(f"[LongMemory][add_record] сохранён текст (первые 100 символов): {text[:100]}")
            logger.debug(f"[LongMemory][add_record] сохранённые метаданные: {metadata}")
            return record_id
        except Exception as e:
            self._log_exception("add_record failed", e)
            return ""

    async def get_all_records(self, filter: Optional[Dict] = None) -> List[dict]:
        logger.debug(f"[LongMemory][get_all_records] вызван")
        try:
            results = self.collection.get(where=filter) if filter else self.collection.get()
            output = [
                {
                    "id": results["ids"][i],
                    "text": results["documents"][i],
                    "metadata": results["metadatas"][i],
                }
                for i in range(len(results["documents"]))
            ]
            logger.debug(f"[LongMemory][get_all_records] вернул {len(output)} записей")
            for rec in output:
                logger.debug(f"[LongMemory][get_all_records] запись: id={rec['id']}, метаданные={rec['metadata']}, текст={rec['text'][:100]}")
            return output
        except Exception as e:
            self._log_exception("get_all_records failed", e)
            return []



import asyncio
import aiosqlite
import logging
from typing import Optional
from datetime import datetime, timedelta

# Настраиваем логгер
logger = logging.getLogger("ShortMemory")
logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
formatter = logging.Formatter('[%(name)s][%(funcName)s] %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

class ShortMemory:
    def __init__(
        self,
        db_path: str = "short_memory.db",        # Путь к SQLite-файлу
        max_pairs: int = 10,                     # Максимум пар user/agent в истории
        ttl_minutes: int = 60,                   # Время хранения записей
        cleanup_interval_minutes: int = 5,       # Частота автоочистки
        start_auto_cleanup: bool = True,
    ):
        self.db_path = db_path
        self.max_pairs = max_pairs
        self.ttl_minutes = ttl_minutes
        self.cleanup_interval_minutes = cleanup_interval_minutes
        self.start_auto_cleanup = start_auto_cleanup
        self._db: Optional[aiosqlite.Connection] = None
        self._initialization_lock = asyncio.Lock()
        logger.debug(f"Инициализирован с db_path={db_path}, max_pairs={max_pairs}, ttl_minutes={ttl_minutes}, cleanup_interval_minutes={cleanup_interval_minutes}, start_auto_cleanup={start_auto_cleanup}")

    def _log_exception(self, message: str, exc: Exception):
        logger.error(f"{message}: {type(exc).__name__} - {exc}")

    async def init(self):
        logger.debug("Вход в init()")
        async with self._initialization_lock:
            if self._db:
                logger.debug("Подключение к БД уже существует, выходим из init")
                return
            try:
                self._db = await aiosqlite.connect(self.db_path)
                logger.debug("Успешное подключение к БД")

                await self._db.execute("""
                    CREATE TABLE IF NOT EXISTS memory (
                        id INTEGER PRIMARY KEY,
                        user_id TEXT,
                        agent_id TEXT,
                        message TEXT,
                        role TEXT,
                        timestamp TEXT
                    );
                """)
                logger.debug("Таблица memory создана или уже существует")

                await self._db.execute("""
                    CREATE INDEX IF NOT EXISTS idx_user_agent_timestamp
                    ON memory (user_id, agent_id, timestamp);
                """)
                logger.debug("Индекс создан или уже существует")

                await self._db.commit()
                logger.debug("Коммит после инициализации выполнен")
            except Exception as e:
                self._log_exception("init failed", e)
                self._db = None

    async def _ensure_db(self):
        logger.debug("Проверка подключения к БД")
        if self._db is None:
            await self.init()
        if self._db is None:
            logger.error("Подключение к базе данных не удалось")
            raise RuntimeError("ShortMemory database is not available.")

    async def close(self):
        logger.debug("Закрытие соединения с БД")
        try:
            if self._db:
                await self._db.close()
                self._db = None
                logger.debug("БД успешно закрыта")
        except Exception as e:
            self._log_exception("close failed", e)

    async def add_message(self, user_id: str, agent_id: str, message: str, role: str):
        logger.debug(f"Добавление сообщения: user_id={user_id}, agent_id={agent_id}, role={role}")
        try:
            await self._ensure_db()
            timestamp = datetime.utcnow().isoformat()
            await self._db.execute(
                "INSERT INTO memory (user_id, agent_id, message, role, timestamp) VALUES (?, ?, ?, ?, ?)",
                (user_id, agent_id, message, role, timestamp)
            )
            logger.debug("Сообщение вставлено")
            await self._db.commit()
            logger.debug("Коммит выполнен")
            await self._enforce_max_pairs(user_id, agent_id)
        except Exception as e:
            self._log_exception("add_message failed", e)

    async def _enforce_max_pairs(self, user_id: str, agent_id: str):
        logger.debug(f"Проверка количества сообщений для user_id={user_id}, agent_id={agent_id}")
        try:
            await self._ensure_db()
            async with self._db.execute("""
                SELECT id FROM memory WHERE user_id = ? AND agent_id = ?
                ORDER BY timestamp DESC
            """, (user_id, agent_id)) as cursor:
                ids = [row[0] async for row in cursor]
            logger.debug(f"Получено {len(ids)} сообщений")
            if len(ids) > self.max_pairs * 2:
                to_delete = ids[self.max_pairs * 2:]
                await self._db.executemany("DELETE FROM memory WHERE id = ?", [(i,) for i in to_delete])
                await self._db.commit()
                logger.debug(f"Удалено {len(to_delete)} старых сообщений")
        except Exception as e:
            self._log_exception("_enforce_max_pairs failed", e)

    async def get_history(self, user_id: str, agent_id: str):
        logger.debug(f"Получение истории сообщений: user_id={user_id}, agent_id={agent_id}")
        try:
            await self._ensure_db()
            async with self._db.execute("""
                SELECT message, role FROM memory
                WHERE user_id = ? AND agent_id = ?
                ORDER BY timestamp ASC
            """, (user_id, agent_id)) as cursor:
                messages = await cursor.fetchall()
            logger.debug(f"Получено {len(messages)} записей")
            pairs = []
            pair = {}
            for msg, role in messages:
                pair[role] = msg
                if "user" in pair and "agent" in pair:
                    pairs.append(pair)
                    pair = {}
            logger.debug(f"Возвращено {len(pairs)} пар сообщений")
            return pairs
        except Exception as e:
            self._log_exception("get_history failed", e)
            return []

    async def cleanup_expired_dialogs(self):
        logger.debug("Очистка устаревших сообщений")
        try:
            await self._ensure_db()
            expiration_threshold = (datetime.utcnow() - timedelta(minutes=self.ttl_minutes)).isoformat()
            await self._db.execute("DELETE FROM memory WHERE timestamp < ?", (expiration_threshold,))
            await self._db.commit()
            logger.debug("Очистка завершена успешно")
        except Exception as e:
            self._log_exception("cleanup_expired_dialogs failed", e)




from typing import Optional, List  # Аннотации типов
from agents import Agent as BaseAgent, Runner, RunConfig  # Импорт базового агента и исполнительной среды
import asyncio
from .memory_manager import MemoryManager  # Импорт менеджера памяти
import uuid
import re
import unicodedata
import logging  # Логгирование

# Настройка логгера
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Преобразование имени агента в безопасный ID
def normalize_agent_id(name: str) -> str:
    ascii_name = unicodedata.normalize("NFKD", name).encode("ascii", "ignore").decode()
    cleaned = re.sub(r'[^a-zA-Z0-9._-]', '_', ascii_name)
    if not cleaned or not re.match(r'^[a-zA-Z0-9]', cleaned):
        cleaned = f"agent_{uuid.uuid4().hex[:8]}"
    return cleaned[:64]  # ограничение длины ID

# Основной класс агента, расширяющий базовый функционал
class Agent(BaseAgent):
    def __init__(
        self,
        name: str,
        *,
        openai_api_key: str = "...",
        ttl_seconds: Optional[int] = None,
        short_memory_max_pairs: int = 10,
        short_memory_ttl_minutes: int = 60,
        short_memory_cleanup_minutes: int = 10,
        max_turns: int = 1,
        **kwargs
    ):
        logger.debug(f"[Agent][__init__] получено имя агента: {name}")
        super().__init__(name=name, **kwargs)

        self.max_turns = max_turns  # Максимальное число reasoning-итераций
        normalized_id = normalize_agent_id(name)  # Преобразуем имя в безопасный ID

        logger.debug(f"[Agent][__init__] нормализованный ID агента: {normalized_id}")

        # Инициализация менеджера памяти
        self.memory_manager = MemoryManager(
            agent_id=normalized_id,
            openai_api_key=openai_api_key,
            long_memory_ttl=ttl_seconds,
            short_memory_max_pairs=short_memory_max_pairs,
            short_memory_ttl_minutes=short_memory_ttl_minutes,
            short_memory_cleanup_minutes=short_memory_cleanup_minutes,
        )

        self._initialized = False  # Флаг инициализации

    async def init(self):
        # Однократная инициализация менеджера памяти
        if self._initialized:
            logger.debug("[Agent][init] уже инициализирован")
            return
        logger.debug("[Agent][init] запускаем инициализацию MemoryManager")
        await self.memory_manager.init()
        self._initialized = True
        logger.debug("[Agent][init] инициализация завершена")

    async def message(
            self,
            input: str,
            *,
            context=None,
            tool_choice: Optional[str] = None,
            max_turns: Optional[int] = None,
            enable_tracing: bool = True,
            tags: Optional[List[str]] = None,
            user_id: Optional[str] = None,
            file_name: Optional[str] = None,
            file_path: Optional[str] = None,
    ):
        logger.debug(f"[Agent][message] получен ввод: {input}")
        await self.init()

        max_turns = max_turns or self.max_turns

        # Получаем память для данного запроса
        memory_context = await self.memory_manager.get_memory_context(user_id=user_id, input=input)

        # 🔍 Логгируем извлечённый memory_context
        if memory_context:
            logger.debug(f"[Agent][message] извлечён memory_context:\n{memory_context}")
        else:
            logger.debug("[Agent][message] память пуста или не извлечена")

        # Формируем prompt с учетом памяти
        if memory_context:
            augmented_input = (
                f"Ты — интеллектуальный агент, ведущий диалог с пользователем с ID: {user_id}.\n"
                f"У тебя есть доступ к двум типам памяти для этого пользователя:\n"
                f"1. 🧠 Краткосрочная память — содержит последние сообщения из диалога.\n"
                f"2. 📚 Долговременная память — содержит знания из прошлых взаимодействий.\n\n"
                f"Используй обе памяти, чтобы понять контекст и дать точный, персонализированный ответ.\n\n"
                f"{memory_context}\n\n"
                f"Теперь ответь на вопрос пользователя:\n{input}"
            )
            logger.debug("[Agent][message] сформирован prompt с памятью")
        else:
            augmented_input = input
            logger.debug("[Agent][message] памяти нет, используется сырой ввод")

        # 🔍 Логгируем финальный prompt
        logger.debug(f"[Agent][message] финальный prompt:\n{augmented_input}")

        # Собираем метаданные
        merged_metadata = {}
        if context and hasattr(context, "metadata") and context.metadata:
            merged_metadata.update(context.metadata)
        for k, v in (("user_id", user_id), ("file_name", file_name), ("file_path", file_path)):
            if v is not None:
                merged_metadata[k] = str(v)

        trace_meta = {}
        if tool_choice is not None:
            merged_metadata["tool_choice"] = tool_choice
            trace_meta["tool_choice"] = tool_choice
        if tags:
            merged_metadata["tags"] = tags
            trace_meta["tags"] = tags

        # Создаем временный объект контекста, если он не передан
        if context is None:
            context = type("Ctx", (), {})()
        context.metadata = merged_metadata or None

        logger.debug(f"[Agent][message] подготовлены метаданные: {merged_metadata}")

        # Конфигурация запуска reasoning-цепочки
        run_cfg = RunConfig(
            tracing_disabled=not enable_tracing,
            trace_metadata=trace_meta if trace_meta else None,
            workflow_name=f"Workflow_{str(user_id)}"
        )

        logger.debug(f"[Agent][message] запускаем Runner с max_turns={max_turns}")

        result = await Runner.run(
            starting_agent=self,
            input=augmented_input,
            context=context,
            max_turns=max_turns,
            run_config=run_cfg,
        )

        logger.debug("[Agent][message] reasoning завершён, сохраняем результат в память")
        try:
            await self.memory_manager.record_interaction(user_id=user_id, input=input, output=result, file_name=file_name, file_path=file_path)
            logger.debug("[Agent][message] Результат успешно сохранён в память, возвращаю результат")
        except Exception as e:
            logger.debug(f"[Agent][message] Не удалось сохранить результат в память, ошибка: {e}")
        return result.final_output